/*
 * Photo Image Print System
 * Copyright (C) 2000-2006 EPSON AVASYS Corporation.
 * Copyright (C) SEIKO EPSON CORPORATION 2000-2006.
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * As a special exception, EPSON AVASYS Corporation gives permission to
 * link the code of this program with libraries which are covered by
 * the EPSON AVASYS Public License and distribute their linked
 * combinations.  You must obey the GNU General Public License in all
 * respects for all of the code used other than the libraries which
 * are covered by EPSON AVASYS Public License.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#ifdef _V1_SUPPORT
#include <string.h>
#endif /* _V1_SUPPORT */

#include "pfpng_ext.h"

#include "pfimg.h"
#include "pfmem.h"
#include "libprtX.h"
#include "pferr.h"
#include "setup.h"


/* Size of buffer for  rgb data storing */
#define DEFAULT_BUFFER_SIZE 0xfffff
/* Size of buffer for pngformat inspection */
#define PNG_CHECK_NUMBER 8

#define WIDTHBYTES(bits) (((bits) + 31) / 32 * 4)
#define SHADING_OFF_FILTER(center, left1, right1, left3, right3) \
((center * 2 + left1 + right1) * 4 + left1 + right1 - left3 - right3) / 16

enum nCONVRSION_STATUS
{
	CONV_STATUS_INIT = 0,
	CONV_STATUS_PAGE_INIT,
	CONV_STATUS_BAND_OUT,
	CONV_STATUS_PAGE_END,
	CONV_STATUS_END,
	CONV_STATUS_PNG_ERROR
};

/* The image information which prints is stored. */
typedef struct _PNG_IMAGE_STRUCT
{
	FILE* ifp;		/*Pointer to a  input stream */
	HANDLE png_ptr;		/* It is used by  libpng. */
	HANDLE info_ptr;	/*It is used by  libpng. */
	int image_pass;		/*Measure against  Interlece */
	POINT image_size;	/* Size of an image */
	POINT print_size;	/* It prints (it actually sends out to library).Size of an image */

	DPOINT magnify;		/* Expansion/reduction rate */
	POINT negative_margin;	/* Minus margin	When a margin is subtracted, it is necessary to cut the picture itself.When a margin is plus, it is 0. */
	int byte_p_pixel;	/* Number of bytes required for  1pixel */
	long read_line;		/* The already read number of rasters */
	int num_line;		/* The storable number of rasters */
	UCHAR** rows;		/* buffer for raster-data storing*/
} IMG_STRUCT, *LP_IMG_STRUCT;


/* Static functions */

/*It initializes by loading  mkcode_init-Library.
The pointer to library is stored in a mkcode structure object.
Argument : Pointer to the share information on LP gas_MKC_STRUCT mkcode
Return value : BOOL A 0 == success and 1== failure */
static BOOL
mkcode_lib_init (LP_MKC_STRUCT);

/*It closes by performing end processing (opening) of  mkcode_end-Library.
Argument : Pointer to the share information on LP gas_MKC_STRUCT mkcode
Return value : BOOL A 0 == success and 1== failure */

static BOOL
mkcode_lib_end (LP_MKC_STRUCT);

/* Initialization of the processing for every  mkcode_page_init-pege
Argument : Pointer to the share information on LP gas_MKC_STRUCT mkcode
Return value : BOOL A 0 == success and 1== failure */
static BOOL
mkcode_page_init (LP_MKC_STRUCT);

/*End processing for every mkcode_page_init-pege
Argument : Pointer to the share information on LP gas_MKC_STRUCT mkcode
Return value : BOOL A 0 == success and 1== failure */
static BOOL
mkcode_page_end (LP_MKC_STRUCT, int);

/* The top function of a  image_to_printer-image conversion loop
It processes in order to all the inputted files.
The pointer to an input stream is stored in a mkcode structure object.
Argument : Pointer to the share information on LP gas_MKC_STRUCT mkcode
Return value : BOOL A 0 == success and 1== failure */
static BOOL
image_to_printer (LP_MKC_STRUCT);

/*As opposed to all the picture information that exists in a  image_convert_to_code-stream,
It processes in order.
Argument : FILE* Pointer to the stream to process
The pointer to the share information on LP gas_MKC_STRUCT mkcode
Return value : BOOL A 0 == success and 1== failure */
static BOOL
image_convert_to_code (FILE*, LP_MKC_STRUCT);

/*The information on a  load_image_init-PNG image is acquired.
An img structure object is built and it stores in a mkcode structure object.
Argument : LP gas_IMG_STRUCT Pointer to the structure object which stores the Image information for processing
The pointer to the share information on LP gas_MKC_STRUCT mkcode
Return value : BOOL A 0 == success and 1== failure */
static BOOL
load_image_init (LP_IMG_STRUCT, LP_MKC_STRUCT);

/*The libpng were using that is ended processing.
A memory is opened. At this time, it is still a stream. It does not close.
Argument : LP gas_IMG_STRUCT Pointer to the structure object which stores the Image information for processing
Return value : void */
static void
load_image_end (LP_IMG_STRUCT);

/*The specification at the time of reading of a  png_setup-PNG image is determined.
If Color becomes here, it will set up so that 24-bit RGB data may be received,
If Monochrome becomes, it will set up so that 8-bit Gray data may be received.
Argument : LP gas_IMG_STRUCT Pointer to the structure object which stores the Image information for processing
char* format at the time of printing is passed.
Return value : BOOL A 0 == success and 1== failure */
static BOOL
png_setup (LP_IMG_STRUCT, char*);


/* load_image_out - It actually processes by library for every raster.
Expansion/reduction processing is also performed here.
Argument : LP gas_IMG_STRUCT Pointer to the structure object which stores the Image information for processing
The pointer to the share information on LP gas_MKC_STRUCT mkcode
Return value : BOOL A 0 == success and 1== failure */
static BOOL
load_image_out (LP_IMG_STRUCT, LP_MKC_STRUCT);

/*The image data for  read_raster_image-1 raster is prepared,
The pointer to the head is returned.
Argument : LP gas_IMG_STRUCT Pointer to the structure object which stores the Image information for processing
Return value : UCHAR* Pointer to the head of data */
static UCHAR*
read_raster_image (LP_IMG_STRUCT);

static void
seek_image (LP_IMG_STRUCT);

/*A setup of  conf_print_image-printing size etc.
Argument : LP gas_PRT_STURCT Printing mode information
LP gas_IMG_STRUCT Pointer to the structure object which stores the Image information for processing
Return value : void */
static void
conf_print_image (LP_PRT_STRUCT, LP_IMG_STRUCT);

/* set_init_param-prt_init is set.
"Argument : Argument passed to LP gas_LIBPRT_INIT DLL_PRT_INIT.
A value is set to this structure object."
LP gas_PRT_STRUCT Printing mode information
char* Output file name (NULL if it becomes standard output)
Return value : void */

static void
set_init_param (ESCPR_OPT*, ESCPR_PRINT_QUALITY*, ESCPR_PRINT_JOB *, LP_PRT_STRUCT);

/* set_band_param-prt_band is set.
"Argument : LP gas_LIBPRT_BAND Argument passed to DLL_BAND_OUT.
A value is set to this structure object."
LP gas_IMG_STRUCT Pointer to the structure object which stores the Image information for processing
Return value : BOOL A 0 == success and 1== failure */
static BOOL
set_band_param (ESCPR_BANDBMP* ,ESCPR_RECT*,LP_IMG_STRUCT);

/*The memory secured at the time of a  free_band_param-set_band_param call is opened.
Argument : LP gas_LIBPRT_BAND Argument passed to DLL_BAND_OUT
Return value : void */
static void
free_band_param (ESCPR_BANDBMP*);

/* write_judge - It writes in from a present processing line and magnifying power, and a good failure is judged.
Argument : long Current line of raster data in preparation
long Current line of the image data under processing
double Magnifying power (%)
Return value : BOOL 0 == write-in improper one and 1== writing are possible. */
static BOOL
write_judge (long, long, double);

#ifdef _V1_SUPPORT
/* escp_through-ESC/P code is skipped.
This exists for [ of v1 ] compatibility and is deletion-planned by v2.
Argument : FILE* Input file pointer
FILE* Output file pointer
char* Pointer to the already read buffer
int Size of the already read buffer
Return value : BOOL -1 == esc/p (not error) and 1== not esc/p */
static BOOL
escp_through (FILE*, FILE*, char*, int);
#endif /* _V1_SUPPORT */

static BOOL
band_flush (void*, /* HANDLE, */ ESCPR_BANDBMP*,ESCPR_RECT*, UCHAR *,/* UCHAR *, */ BOOL);

static void
set_raster (UCHAR *, UCHAR *, int, int, double, int);


static ESCPR_BYTE4
print_spool_fnc(void* , const ESCPR_UBYTE1*, ESCPR_UBYTE4);

FILE* outfp = NULL;

/**************************************************
* public functions
**************************************************/
/*image is changed into ESC/P by accessing  make_code_process-libPrtXXX.
Argument : LP gas_MKC_STRUCT Path to a library required for this structure object
The inputted list of files, and a printing parameter

Return value : void */
BOOL
make_code_process (LP_MKC_STRUCT lp_mkcode)
{
	int eflag = 0;

	/* An output place is opened. */
	if (lp_mkcode->output_file == NULL)
	{
		outfp = stdout;
	}
	else
	{
		/*It makes it possible to output to _LPR_DIRECT-lpr directly. Schedule deleted in version2 */

#ifdef _LPR_DIRECT
		if (strcmp (lp_mkcode->output_file, "."))
		{
#endif /* _LPR_DIRECT */
			outfp = fopen (lp_mkcode->output_file, "wb");
			if (outfp == NULL)
			{
				PF_ERROR_HANDLER (FILE_NOT_FOUND, lp_mkcode->output_file);
				return 1;
			}
#ifdef _LPR_DIRECT
		}
		else
		{
			char prt_name[107];

			strcpy (prt_name, "lpr -P");
			get_prt_name (prt_name + 6);
			outfp = popen (prt_name, "w");
			if (outfp == NULL)
				fprintf (stderr, _("Not able to communicate with lpr.\n"));
		}
#endif /* _LPR_DIRECT */
	}

	lp_mkcode->output_fp = outfp;

	lp_mkcode->conv_status = 0;

	/* Conversion process */
	eflag = image_to_printer (lp_mkcode);

	/* End processing */
	if (lp_mkcode->conv_status)
		eflag |= mkcode_lib_end (lp_mkcode);

	if (lp_mkcode->output_file != NULL)
	{
#ifdef _LPR_DIRECT
		if (strcmp (lp_mkcode->output_file, "."))
		{
#endif /* not _LPR_DIRECT */
			fclose (outfp);
#ifdef _LPR_DIRECT
		}
		else
		{
			pclose (outfp);
		}
#endif /* _LPR_DIRECT */
	}
	if (eflag)
	{
		/* PF_ERROR_HANDLER (ABNORMAL_END, NULL); */
		return 1;
	}

	return 0;
}

/**************************************************
* static functions
**************************************************/
static BOOL
mkcode_lib_init (LP_MKC_STRUCT lp_mkcode)
{
	lt_dlhandle lib_handle = NULL;
	char *lib_path = lp_mkcode->libprt_path;
	INIT_FUNC func;
	ESCPR_OPT RGB_opt;
	ESCPR_PRINT_QUALITY PrintQuality;
	ESCPR_PRINT_JOB PrintJob;
	int err = 0;

	/*library specified by  lib_path is loaded. */
	err = lt_dlinit();
	if (!err) {
		lib_handle = lt_dlopen (lib_path);
	}
	if (lib_handle == NULL)
	{
		PF_ERROR_HANDLER (DLL_ERROR, (char*)lt_dlerror());
		return 1;
	}
	lp_mkcode->lib_handle = lib_handle;

	/* Initialization function call of  library */
	func = (INIT_FUNC)lt_dlsym (lib_handle, DLL_PRT_INIT);
	if (func == NULL)
	{
		PF_ERROR_HANDLER (DLL_ERROR, (char*)lt_dlerror());
		return 1;
	}

	/* paramater set */
	set_init_param (&RGB_opt, &PrintQuality, &PrintJob, lp_mkcode->lp_prtinfo);

	/* Initialization of  library */
	if((*func)(&RGB_opt, &PrintQuality, &PrintJob))
	{
		PF_ERROR_HANDLER (LIBPRT_ERROR, lib_path);
		return 1;
	}

	lp_mkcode->conv_status = CONV_STATUS_INIT;
	return 0;
}

static BOOL
mkcode_lib_end (LP_MKC_STRUCT lp_mkcode)
{
	HANDLE lib_handle = lp_mkcode->lib_handle;
	END_FUNC func;

	if (lib_handle == NULL)
		return 1;

	/*  End processing function call of library */
	func = (END_FUNC)lt_dlsym(lib_handle, DLL_PRT_END);
	if (func == NULL)
	{
		PF_ERROR_HANDLER (DLL_ERROR, (char*)lt_dlerror ());
		return 1;
	}

	/* Checking library? */
	if ((*func)(/* prt_handle */))
	{
		PF_ERROR_HANDLER (LIBPRT_ERROR, lp_mkcode->libprt_path);
		return 1;
	}

	/* Closing of  library */
	if (lt_dlclose (lib_handle) == -1)
	{
		PF_ERROR_HANDLER (DLL_ERROR, (char*)lt_dlerror ());
		return 1;
	}
	lt_dlexit();

	lp_mkcode->conv_status = CONV_STATUS_END;
	return 0;
}


static BOOL
mkcode_page_init (LP_MKC_STRUCT lp_mkcode)
{
	HANDLE lib_handle = lp_mkcode->lib_handle;
	PINIT_FUNC func;

	func = (PINIT_FUNC)lt_dlsym (lib_handle, DLL_PAGE_INIT);
	if (func == NULL)
	{
		PF_ERROR_HANDLER (DLL_ERROR, (char*)lt_dlerror ());
		return 1;
	}

	if ((*func) ())
	{
		PF_ERROR_HANDLER (LIBPRT_ERROR, lp_mkcode->libprt_path);
		return 1;
	}
	lp_mkcode->conv_status = CONV_STATUS_PAGE_INIT;
	return 0;
}


static BOOL
mkcode_page_end (LP_MKC_STRUCT lp_mkcode, int flag)
{
	HANDLE lib_handle = lp_mkcode->lib_handle;
	PEND_FUNC func;

	func = (PEND_FUNC)lt_dlsym (lib_handle, DLL_PAGE_END);
	if (func == NULL)
	{
		PF_ERROR_HANDLER (DLL_ERROR, (char*)lt_dlerror());
		return 1;
	}

	if ((*func) (flag))
	{
		PF_ERROR_HANDLER (LIBPRT_ERROR, lp_mkcode->libprt_path);
		return 1;
	}

	lp_mkcode->conv_status = CONV_STATUS_PAGE_END;
	return 0;
}

/* Printing job loop */
static BOOL
image_to_printer (LP_MKC_STRUCT lp_mkcode)
{
	char** input_files = lp_mkcode->input_files;
	BOOL eflag = 0;
	int in_flag, count;

	in_flag = count = 0;
	/* It is a loop until it processes all the inputted files. */
	while (!eflag && !in_flag)
	{
		FILE *fp = NULL;

		/* From standard input */
		if (input_files == NULL)
		{
			fp = stdin;
			in_flag = 1;
		}
		/* From a file */
		else
		{
			fp = fopen (input_files[count], "rb");
			if (fp == NULL)
			{
				PF_ERROR_HANDLER (FILE_NOT_FOUND, input_files[count]);
				return 1;
			}
		}

		eflag = image_convert_to_code (fp, lp_mkcode);

		if (!in_flag)
		{
			fclose(fp);
			if (eflag)
				PF_ERROR_HANDLER (CONVERT_ERROR, input_files[count]);

			if (input_files[++count] == NULL) in_flag = 1;
		}
		else if (eflag)
		{
			PF_ERROR_HANDLER (CONVERT_ERROR, NULL);
		}
	}
	return eflag;
}

static BOOL
image_convert_to_code (FILE* fp, LP_MKC_STRUCT lp_mkcode)
{
	IMG_STRUCT image_str;
	BOOL loop_flag;
	BOOL eflag = 0;

	image_str.ifp = fp;
	/* It is a loop in order to process all the files under present processing. */
	loop_flag = 1;
	while (loop_flag)
	{
		int stat = 0;
		/* Picture information is read. */
		stat = load_image_init (&image_str,
					lp_mkcode);
		switch (stat)
		{
		case 1:		/* reading error */
			eflag = 1;

		case -1:		/* Processing end */
			loop_flag = 0;
			continue;

		case 0:			/* The completion of picture taking-in preparation */
			/* library is loaded -- initialization */
			if (!lp_mkcode->conv_status)
			{
				if (mkcode_lib_init (lp_mkcode))
					return 1;
				lp_mkcode->conv_status = 1;
			}
		default: break;
		}

		/* Page initialization processing */
		eflag = mkcode_page_init (lp_mkcode);
		if (eflag)
			break;
		/* ESC/P code beginning */
		eflag = load_image_out (&image_str, lp_mkcode);

		if (lp_mkcode->conv_status != CONV_STATUS_PNG_ERROR)
			load_image_end (&image_str);

		/* Page end processing */
		mkcode_page_end (lp_mkcode, eflag);

		if (eflag)
			break;
	}

	return eflag;
}

/* Information is acquired from the inputted PNG file. */
static BOOL
load_image_init (LP_IMG_STRUCT lp_image, LP_MKC_STRUCT lp_mkcode)
{
	FILE* fp = lp_image->ifp;
	char header[PNG_CHECK_NUMBER];

	png_uint_32 ver;
	char version [32];
	png_structp png_ptr;
	png_infop info_ptr;

	/* Dynamic linking to PNG library. */
	if (!open_png_library ())
		return 1;

	ver = dl_png_access_version_number ( );

	sprintf ( version, "%ld.%ld.%ld",
		ver / 10000,
		( ver / 100 ) % 100,
		ver % 100 );

	if (fread (header, sizeof (char), PNG_CHECK_NUMBER, fp)
	    != PNG_CHECK_NUMBER)
		return -1;		/* With no input (a processing end is expressed) */
	 /* Check of format */
	if (dl_png_sig_cmp ((png_bytep)header, (png_size_t)0, PNG_CHECK_NUMBER))
	{
#ifdef _V1_SUPPORT
		 /* ESC/P code is skipped.
		This exists for [ of v1 ] compatibility and is a deletion schedule at v2. */
		return escp_through (fp, lp_mkcode->output_fp,
				header, PNG_CHECK_NUMBER);
#else
		return 1;
#endif /* _V1_SUPPORT */
	}
	/* The inside information (handle) of  libpng is obtained.
	The default error handling of libpng is used. */
	png_ptr = dl_png_create_read_struct (version,
					NULL, NULL, NULL);
	if (!png_ptr)
		return 1;

	lp_image->png_ptr = (void*)png_ptr;

	/* info structure object is created.
	It is safer for access to this structure object to use the function of libpng. */
	info_ptr = dl_png_create_info_struct (png_ptr);
	if (!info_ptr)
		return 1;

	lp_image->info_ptr = (void*)info_ptr;

	/* setjmp of a sake when an error takes place within libpng */
	if (setjmp (png_ptr->jmpbuf))
	{
		PF_ERROR_HANDLER (LIB_PNG_ERROR, NULL);
		lp_mkcode->conv_status = CONV_STATUS_PNG_ERROR;
		return 1;
	}
	/* png is processed. */
	if (lp_mkcode->lp_prtinfo->ink)
		return png_setup (lp_image, "MONO");
	else
		return png_setup (lp_image, "COLOR");
}

static void
load_image_end (LP_IMG_STRUCT lp_image)
{
	int i;

	if (lp_image == NULL) return;
	if (lp_image->png_ptr || lp_image->info_ptr)
	{
		png_structp png_ptr = (png_structp)lp_image->png_ptr;
		png_infop info_ptr = (png_infop)lp_image->info_ptr;

		dl_png_read_end(png_ptr, info_ptr);
		dl_png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
	}
	for (i = 0; i < lp_image->num_line; i++)
		pf_mem_free (lp_image->rows[i]);

	pf_mem_free (lp_image->rows);

	/* Unlinking from PNG library. */
	close_png_library ();

	return;
}

/* A setup of png_setup-png */

static BOOL
png_setup (LP_IMG_STRUCT lp_image, char* format)
{
	UCHAR** rows;
	int line, i;
	long width_byte;
	png_uint_32 width, height;
	int bit_depth, color_type;

	png_structp png_ptr = lp_image->png_ptr;
	png_infop info_ptr = lp_image->info_ptr;
	FILE* fp = lp_image->ifp;

	/* I/O initialization */
	dl_png_init_io (png_ptr, fp);
	/* The already read portion is taught. */
	dl_png_set_sig_bytes (png_ptr, PNG_CHECK_NUMBER);

	/* The information on  png is acquired. */
	dl_png_read_info (png_ptr, info_ptr);

	/* IHDR reading */
	dl_png_get_IHDR (png_ptr, info_ptr, &width, &height,
		&bit_depth, &color_type,
		NULL, NULL, NULL);
	lp_image->image_size.x = width;
	lp_image->image_size.y = height;

	/* Interlace */
	if (dl_png_set_interlace_handling (png_ptr) != 1)
	{
		PF_ERROR_HANDLER (NO_SUPPORT_ERROR, "\"Interlace\" is not supported.");
		return 1;
	}

	/* Color data is set as 8 bits. */
	if ((color_type == PNG_COLOR_TYPE_PALETTE && bit_depth <= 8)
	|| (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
	|| (dl_png_get_valid (png_ptr, info_ptr, PNG_INFO_tRNS)))
		dl_png_set_expand (png_ptr);

	if (bit_depth < 8)
		dl_png_set_packing (png_ptr);

	if (bit_depth == 16)
		dl_png_set_strip_16 (png_ptr);

	if (color_type == PNG_COLOR_TYPE_RGB_ALPHA)
		dl_png_set_strip_alpha(png_ptr);

	/*In  color printing, in the case of monochrome, it is set as 1 byte at 3byteBGR. */
	if (!strcmp (format, "MONO"))	/* Monochrome */
	{
		if (color_type == PNG_COLOR_TYPE_RGB
		|| color_type == PNG_COLOR_TYPE_RGB_ALPHA
		|| color_type == PNG_COLOR_TYPE_PALETTE)
		{
			dl_png_set_rgb_to_gray (png_ptr, 1, 0.3, 0.59);

			/* Convert data to 24bits monochrome */
			dl_png_set_gray_to_rgb (png_ptr);
		}
		width_byte = WIDTHBYTES(width * 24/* 8 */);
		line = DEFAULT_BUFFER_SIZE / width_byte;

	}
	else	/* Color */
	{
		if (color_type == PNG_COLOR_TYPE_GRAY
		|| color_type == PNG_COLOR_TYPE_GRAY_ALPHA
		|| color_type == PNG_COLOR_TYPE_PALETTE)
			dl_png_set_gray_to_rgb (png_ptr);
		width_byte = WIDTHBYTES(width * 24);
		line = DEFAULT_BUFFER_SIZE / width_byte;
	}

	/* A setup of  former is made to reflect. */
	dl_png_read_update_info (png_ptr, info_ptr);

	/* buffer for raster-data storing is secured per party.
	The maximum size which does not exceed DEFAULT_BUFFER_SIZE is secured. */
	rows = (UCHAR**) pf_mem_alloc (sizeof (png_bytep) * line);
	if (rows == NULL)
	{
		PF_ERROR_HANDLER (MEMORY_ERROR, NULL);
		return 1;
	}
	lp_image->rows = rows;

	line = (line > 0) ? line : 1;

	for (i = 0; i < line; i++)
	{
		rows[i] = pf_mem_alloc (width_byte);

		if (rows[i] == NULL)
		{
			PF_ERROR_HANDLER (MEMORY_ERROR, NULL);
			return 1;
		}
	}

	lp_image->num_line = line;
	return 0;
}


static BOOL
load_image_out (LP_IMG_STRUCT lp_image, LP_MKC_STRUCT lp_mkcode)
{
	ESCPR_BANDBMP InBmp;
	ESCPR_RECT BandRec;

	long h_ofs;
	UCHAR *raster;
	void* func;
	long h_prt_count, h_img_count;
	double h_mag;

	conf_print_image (lp_mkcode->lp_prtinfo, lp_image);

	func = (void *)lt_dlsym (lp_mkcode->lib_handle, DLL_BAND_OUT);
	if (set_band_param (&InBmp, &BandRec, lp_image))
		return 1;

	if (func == NULL)
	{
		PF_ERROR_HANDLER (DLL_ERROR, (char*)lt_dlerror ());
		return 1;
	}

	raster = (UCHAR*)pf_mem_alloc (InBmp.WidthBytes + 3);
	if (!raster)
	{
		PF_ERROR_HANDLER (MEMORY_ERROR, NULL);
		return 1;
	}

	memset ((UCHAR*)raster, 0xff, InBmp.WidthBytes + 3);

	h_prt_count = lp_image->print_size.y + lp_image->negative_margin.y;
	h_img_count = 0;
	h_mag = lp_image->magnify.y;

	for (h_ofs = 0; h_prt_count > 0; h_ofs ++)
	{
		UCHAR *image;

		/* Acquisition of  image */
		image = read_raster_image (lp_image);

		if (!image)
		{
			PF_ERROR_HANDLER (UNEXPECTED_ERROR, NULL);
			return 1;
		}

		set_raster (raster, image, lp_image->print_size.x, lp_image->negative_margin.x,
			lp_image->magnify.x, lp_image->byte_p_pixel);

		/* Lengthwise direction */
		while (h_prt_count > 0 && write_judge (h_img_count, h_ofs, h_mag))
		{
			if (h_prt_count <= lp_image->print_size.y)
			{
				if (band_flush (func, &InBmp, &BandRec, raster, 0))
				{
					PF_ERROR_HANDLER (LIBPRT_ERROR, lp_mkcode->libprt_path);
					return 1;
				}
			}
			h_img_count ++;
			h_prt_count --;
		}
	}

	if (band_flush (func, &InBmp, &BandRec, NULL,1))
	{
		PF_ERROR_HANDLER (LIBPRT_ERROR, lp_mkcode->libprt_path);
		return 1;
	}

	seek_image (lp_image);

	free_band_param (&InBmp);
	lp_mkcode->conv_status = CONV_STATUS_BAND_OUT;

	return 0;
}


static UCHAR*
read_raster_image (LP_IMG_STRUCT lp_image)
{
	long read_count = lp_image->read_line;
	int line_ofs;

	if (lp_image->read_line >= lp_image->image_size.y)
		return NULL;

	if (read_count == 0 || read_count % lp_image->num_line == 0)
	{
		/* Data is read from  png and the pointer to the first raster is returned. */
		png_structp png_ptr = (png_structp)lp_image->png_ptr;
		long line;

		/* The number of rasters to read (it can read) is determined. */
		if (lp_image->num_line + read_count < lp_image->image_size.y)
			line = lp_image->num_line;
		else
			line = lp_image->image_size.y - read_count;


		dl_png_read_rows (png_ptr, lp_image->rows, NULL, line);
		line_ofs = 0;
	}
	else
	{
		/* The pointer to the already read raster is returned. */
		line_ofs = read_count % lp_image->num_line;
	}

	lp_image->read_line ++;
	return lp_image->rows[line_ofs];
}

static void
seek_image (LP_IMG_STRUCT lp_image)
{
	long line_number = lp_image->image_size.y - lp_image->read_line;
	long i;

	for (i = 0; i < line_number; i ++)
	{
		read_raster_image (lp_image);
	}
	return;
}

static void
conf_print_image (LP_PRT_STRUCT lp_prtinfo, LP_IMG_STRUCT lp_image)
{
	long positive_margin_x, positive_margin_y;
	double area_y;		/* For the operation in 3mm printing of lower ends */

	area_y = lp_prtinfo->paper_area.y;
	lp_image->byte_p_pixel = 3;

	if (lp_prtinfo->fit_page)
	{
		lp_image->negative_margin.x = 0;
		lp_image->negative_margin.y = 0;
		positive_margin_x = 0;
		positive_margin_y = 0;
		lp_image->magnify.x = (double)lp_prtinfo->paper_area.x / (double)lp_image->image_size.x;
		lp_image->magnify.y = area_y / (double)lp_image->image_size.y;

		if (lp_image->magnify.x < lp_image->magnify.y)
		{
			lp_image->magnify.y = lp_image->magnify.x;
		}
		else
		{
			lp_image->magnify.x = lp_image->magnify.y;
		}
	}
	else
	{
		/* Setup of a margin */
		if (lp_prtinfo->margin.x < lp_prtinfo->default_margin.x)
		{
			lp_image->negative_margin.x = fabs ((double)lp_prtinfo->margin.x
							- lp_prtinfo->default_margin.x);
			positive_margin_x = 0;
		}
		else
		{
			lp_image->negative_margin.x = 0;
			positive_margin_x = lp_prtinfo->margin.x - lp_prtinfo->default_margin.x;
		}

		if (lp_prtinfo->margin.y < lp_prtinfo->default_margin.y)
		{
			lp_image->negative_margin.y = fabs ((double)lp_prtinfo->margin.y
							- lp_prtinfo->default_margin.y);
			positive_margin_y = 0;
		}
		else
		{
			lp_image->negative_margin.y = 0;
			positive_margin_y = lp_prtinfo->margin.y - lp_prtinfo->default_margin.y;
		}

		/* Expansion/reduction */
		lp_image->magnify.x = (double)lp_prtinfo->ratio.x / 100.0;
		lp_image->magnify.y = (double)lp_prtinfo->ratio.y / 100.0;
	}

	if (lp_prtinfo->borderless_print == 1)
	{
		double t_mag;
		double ty_mag;

		t_mag = (double)lp_prtinfo->paper_area.x / lp_prtinfo->paper_size.x;
		ty_mag = (double)lp_prtinfo->paper_area.y / lp_prtinfo->paper_size.y;
		t_mag = (t_mag > ty_mag) ? t_mag : ty_mag;

		lp_image->magnify.x *= t_mag;
		lp_image->magnify.y *= t_mag;
		t_mag = (double)lp_prtinfo->paper_area.x / lp_prtinfo->paper_size.x;

	}


	/* Printing domain */
	if((lp_prtinfo->paper_area.x - positive_margin_x)
	< ((lp_image->image_size.x - lp_image->negative_margin.x) * lp_image->magnify.x))
	{
		lp_image->print_size.x = lp_prtinfo->paper_area.x - positive_margin_x;
	}
	else
	{
		lp_image->print_size.x = (lp_image->image_size.x * lp_image->magnify.x)
			- lp_image->negative_margin.x;
	}

	if (lp_image->print_size.x % 2)
		lp_image->print_size.x--;

	if((area_y - positive_margin_y)
	< ((lp_image->image_size.y - lp_image->negative_margin.y) * lp_image->magnify.y))
	{
		lp_image->print_size.y = area_y - positive_margin_y;
	}
	else
	{
		lp_image->print_size.y = (lp_image->image_size.y * lp_image->magnify.y)
			- lp_image->negative_margin.y;
	}


	lp_image->read_line = 0;
}

static void
set_init_param (ESCPR_OPT *pRGB_opt, ESCPR_PRINT_QUALITY *pPrintQuality,
		ESCPR_PRINT_JOB *pPrintJob, LP_PRT_STRUCT lp_prtinfo)
{

	pRGB_opt->fpspoolfunc = (ESCPR_FPSPOOLFUNC)print_spool_fnc;

	pPrintQuality->MediaTypeID = lp_prtinfo->media_type;
	pPrintQuality->PrintQuality = lp_prtinfo->qlevel;
	pPrintQuality->ColorMono = lp_prtinfo->ink;
	pPrintQuality->Brightness = lp_prtinfo->brightness;
	pPrintQuality->Contrast = lp_prtinfo->contrast;
	pPrintQuality->Saturation = lp_prtinfo->saturation;
	pPrintQuality->ColorPlane = ESCPR_CP_FULLCOLOR;
	pPrintQuality->PaletteSize = 0;
	pPrintQuality->PaletteData = NULL;

	pPrintJob->PaperWidth = lp_prtinfo->paper_size.x;
	pPrintJob->PaperLength = lp_prtinfo->paper_size.y;

	pPrintJob->TopMargin = lp_prtinfo->margin.y;
	pPrintJob->LeftMargin = lp_prtinfo->margin.x;

	if (lp_prtinfo->borderless_print == 1)
	{
		if (pPrintJob->PaperWidth < 4209)
			pPrintJob->PrintableAreaWidth = pPrintJob->PaperWidth + 72; /* 36 + 36 */
		else
			pPrintJob->PrintableAreaWidth = pPrintJob->PaperWidth + 96; /* 48 + 48 */

		pPrintJob->PrintableAreaLength = pPrintJob->PaperLength + 112; /* 42 + 70 */

	}
	else
	{
		if (pPrintJob->TopMargin < 42)
			pPrintJob->TopMargin = 42;

		if (pPrintJob->LeftMargin < 42)
			pPrintJob->LeftMargin = 42;

		pPrintJob->PrintableAreaWidth        = (lp_prtinfo->margin.x < lp_prtinfo->default_margin.x)
			? lp_prtinfo->paper_area.x
			: lp_prtinfo->paper_area.x - lp_prtinfo->margin.x + lp_prtinfo->default_margin.x;

		pPrintJob->PrintableAreaLength       = (lp_prtinfo->margin.y < lp_prtinfo->default_margin.y)
			? lp_prtinfo->paper_area.y
			: lp_prtinfo->paper_area.y - lp_prtinfo->margin.y + lp_prtinfo->default_margin.y;
	}
	pPrintJob->InResolution = ESCPR_IR_3636;
	pPrintJob->PrintDirection = ESCPR_PD_BIDIREC;

#if DEBUG
	fprintf (stderr,
		"**************************************************\n"
		 "                    DEBUG DUMP                    \n"
		 "**************************************************\n"
		 "\n[PrintQuality]\n" \
		 "MediaTypeID  = %d\n" \
		 "PrintQuality = %d\n" \
		 "ColorMono    = %d\n" \
		 "Brightness   = %d\n" \
		 "Contrast     = %d\n" \
		 "Saturation   = %d\n" \
		 "ColorPlane   = %d\n" \
		 "PaletteSize  = %d\n" \
		 "PaletteData  = %d\n" \
		 "\n[pPrintJob]\n" \
		"PaperWidth = %d\n" \
		"PaperLength = %d\n" \
		"TopMargin = %d\n" \
		"LeftMargin = %d\n" \
		"PrintableAreaWidth = %d\n" \
		"PrintableAreaLength = %d\n" \
		"InResolution = %d\n" \
		"PrintDirection = %d\n" \
		"\n**************************************************\n\n",
		pPrintQuality->MediaTypeID,
		pPrintQuality->PrintQuality,
		pPrintQuality->ColorMono,
		pPrintQuality->Brightness,
		pPrintQuality->Contrast,
		pPrintQuality->Saturation,
		pPrintQuality->ColorPlane,
		pPrintQuality->PaletteSize,
		pPrintQuality->PaletteData,
		pPrintJob->PaperWidth,
		pPrintJob->PaperLength,
		pPrintJob->TopMargin,
		pPrintJob->LeftMargin,
		pPrintJob->PrintableAreaWidth,
		pPrintJob->PrintableAreaLength,
		pPrintJob->InResolution,
		pPrintJob->PrintDirection);
#endif
	return;
}

static BOOL
set_band_param (ESCPR_BANDBMP *pInBmp,ESCPR_RECT *pBandRec,LP_IMG_STRUCT lp_image)
{
	UCHAR* ex_raster;
	long width_pixel;
	long width_bytes;

	width_pixel = lp_image->print_size.x;
	width_bytes = WIDTHBYTES(width_pixel * lp_image->byte_p_pixel * 8);

	ex_raster = (UCHAR*)pf_mem_alloc (width_bytes * 2);
	if (!ex_raster)
	{
		PF_ERROR_HANDLER (MEMORY_ERROR, NULL);
		return 1;
	}

	pBandRec->top = 0;
	pBandRec->bottom = 0;
	pBandRec->left = 0;
	pBandRec->right = width_pixel;
	pInBmp->WidthBytes = width_bytes;
	pInBmp->Bits = ex_raster;

	return 0;
}

static void
free_band_param (ESCPR_BANDBMP* pInBmp)

{
	if (pInBmp->Bits != NULL)
		pf_mem_free (pInBmp->Bits);
	return;
}

static BOOL
write_judge (long count, long point, double mag)
{
	long mag_count = floor ((point + 1) * mag);
	long now_count = count + 1;

	if (mag_count < now_count)
		return 0;			/* write-in disapproval */

	/* write-in permission */
	return 1;
}

#ifdef _V1_SUPPORT
/* ESC/P code is skipped.
This exists for [ of v1 ] compatibility and is a deletion schedule at v2. */
static BOOL
escp_through (FILE* infp, FILE* outfp, char* rbuf, int rnum)
{
	char escp_head[] = "REMOTE1";
	char buffer[DEFAULT_BUFFER_SIZE];
	long numbuf = (long)rnum;
	long i;

	if (rbuf == NULL || rnum == 0)
		return 1;

	memcpy (buffer, rbuf, rnum);
	numbuf += fread (buffer + rnum, sizeof (char), DEFAULT_BUFFER_SIZE - rnum, infp);
	if ( numbuf < strlen (escp_head))
		return 1;

	for (i = 0; i < numbuf - strlen (escp_head); i++)
	{
		if (!strncmp (buffer + i, escp_head, strlen (escp_head)))
		{
			fwrite (buffer, sizeof (char), numbuf, outfp);
			while ((numbuf = fread (buffer, sizeof (char), DEFAULT_BUFFER_SIZE, infp))
			!= 0)
			{
				fwrite (buffer, sizeof (char), numbuf, outfp);
				fflush (outfp);
			}
			if (numbuf) fwrite (buffer, sizeof (char), numbuf, outfp );
			return -1;
		}
	}
	return 1;
}
#endif /* _V1_SUPPORT */


/* 2 line For delivery flag */
BOOL _2_Line_Flag;


static BOOL
band_flush (void* func, ESCPR_BANDBMP* pInBmp, ESCPR_RECT* pBandRec,
	UCHAR *raster, BOOL flush_flag)

{
	UCHAR *target_line;
	if (flush_flag)
	{
		if (!_2_Line_Flag)
			return 0;
		pBandRec->bottom = pBandRec->top+ 1;
	}
	else
	{
		pBandRec->bottom = pBandRec->top+ 2;

		if (_2_Line_Flag)
			target_line = pInBmp->Bits + pInBmp->WidthBytes;
		else
			target_line = pInBmp->Bits;

		memcpy (target_line, raster, pInBmp->WidthBytes);
	}

	if (_2_Line_Flag)
	{
		if ((*(OUT_FUNC)func) (pInBmp, pBandRec))
		{
			return 1;
		}
		pBandRec->top = pBandRec->bottom;
	}

	_2_Line_Flag ^= 1;
	return 0;
}


static void
set_raster (UCHAR* raster, UCHAR* image_band, int size, int negative, double mag, int bpp)
{
	int w_prt_count = size + negative;
	double w_mag = mag;
	int w_img_count = 0;
	int w_ofs;

	for (w_ofs = 0; w_prt_count > 0; w_ofs++)
	{
		while (w_prt_count > 0 && write_judge (w_img_count, w_ofs, w_mag))
		{
			int i;
			UCHAR* to_band;

			if (w_prt_count <= size)
			{
				to_band = &(raster[(w_img_count - negative) * bpp]);

				for (i = 0; i < bpp; i++)
					to_band[i] = image_band[w_ofs * bpp + i];
			}
			w_img_count ++;
			w_prt_count --;
		}
	}
	return;
}

ESCPR_BYTE4
print_spool_fnc(void* hParam, const ESCPR_UBYTE1* pBuf, ESCPR_UBYTE4 cbBuf)
{
	long int i;
	for (i = 0; i < cbBuf; i++)
		putc(*(pBuf + i), outfp);
	return 1;

}
